import 'dart:convert';
import 'dart:io';

import 'package:build/build.dart';
import 'package:oobium_datastore_gen/src/feature.dart';
import 'package:oobium_datastore_gen/src/generators/console_feature.dart';
import 'package:oobium_datastore_gen/src/generators/persist_feature.dart';
import 'package:oobium_datastore_gen/src/generators/rest_feature.dart';
import 'package:oobium_datastore_gen/src/generators/library.dart';
import 'package:oobium_datastore_gen/src/schema_parser.dart';
import 'package:oobium_datastore_gen/src/util.dart';
import 'package:xstring/xstring.dart';

Builder schemaBuilder(BuilderOptions options) => SchemaBuilder();

class SchemaBuilder implements Builder {
  @override
  Future build(BuildStep buildStep) async {
    final inputId = buildStep.inputId;
    banner('running: SchemaBuilder on $inputId');

    final xFile = File(inputId.addExtension('.x.dart').path);

    final parts = inputId.pathSegments.last.split('.');
    final name = parts[0];
    final version = (parts[1] != 'schema') ? parts[1] : null;
    final contents = await buildStep.readAsString(inputId);

    final lines = LineSplitter.split(contents).map((line) {
      return line.replaceAllMapped(RegExp(r'{{\s*(.*?)\s*}}'), (m) {
        final name = m.group(1)!;
        final value = Platform.environment[name];
        if(value == null) {
          print('environment variable "$name" is not set');
          return m.group(0)!;
        } else {
          print('"$name" -> "$value"');
          return value;
        }
      });
    });

    final schema = SchemaParser().parse(name, version, lines);

    final xSource = read(xFile);
    final sources = <String>[];
    for(final import in schema.imports) {
      final assetId = AssetId(import.package, import.path);
      final unit = await buildStep.resolver.compilationUnitFor(assetId);
      sources.add(unit.toSource());
    }

    schema.isEnum = (type) => [xSource, ...sources].any((s) => s.containsEnum(type));
    schema.isInPart = (type) => xSource.containsClass(type);

    final genId = inputId.addExtension('.g.dart');
    final genSrc = genLibrary(schema);
    await buildStep.writeAsString(genId, tryFormat(genSrc));

    if(schema.has<ConsoleFeature>()) {
      final id = inputId.addExtension('.g.feature_console.dart');
      final src = genConsole(schema);
      await buildStep.writeAsString(id, tryFormat(src));
    }

    if(schema.has<PersistFeature>()) {
      final id = inputId.addExtension('.g.feature_persist.dart');
      final src = genPersist(schema);
      await buildStep.writeAsString(id, tryFormat(src));
    }

    if(schema.has<RestFeature>()) {
      final id = inputId.addExtension('.g.feature_rest.dart');
      final src = genRest(schema);
      await buildStep.writeAsString(id, tryFormat(src));
    }

    write(xFile.path, schema.partOfDeclaration, overwrite: false);
  }

  @override
  final buildExtensions = const {
    '.schema': [
      '.schema.g.dart',
      '.schema.g.feature_console.dart',
      '.schema.g.feature_persist.dart',
      '.schema.g.feature_rest.dart',
    ]
  };
}

extension StringX on String {
  bool containsClass(String type) {
    return contains(RegExp('\\s*class\\s+$type(?:\\s|\\{)'));
  }
  bool containsEnum(String type) {
    return contains(RegExp('\\s*enum\\s+$type(?:\\s|\\{)'));
  }
}