import 'package:collection/collection.dart';
import 'package:oobium_datastore_gen/src/class.dart';

class ClassField {

  final String name;
  final String type;
  final bool _nullable;
  final String? initializer;
  final metadata = <String>[];

  ClassField({
    required this.name,
    required this.type,
    bool nullable = false,
    String? initializer,
    Iterable<String>? metadata
  }) :
    _nullable = nullable,
    initializer = initializer {
    metadata?.forEach((m) => this.metadata.add(m));
  }
  ClassField.paramOf(ClassField f) :
        parent = f.parent, type = f.parameter, name ='', _nullable = false, initializer = null;

  late final Class parent;

  String get nullable => isNullable ? '?' : '';

  bool get isDateTime => type == 'DateTime';
  bool get isNotDateTime => !isDateTime;
  bool get isNullable => _nullable;
  bool get isNotNullable => !isNullable;
  bool get isString => type == 'String';
  bool get isNotString => !isString;
  bool get isIterable => type == 'Iterable' || type.startsWith('Iterable<');
  bool get isNotIterable => !isIterable;
  bool get isCollection => isList || isIterable || isSet || isMap;
  bool get isNotCollection => !isCollection;
  bool get isList => type == 'List' || type.startsWith('List<');
  bool get isNotList => !isList;
  bool get isSet => type == 'Set' || type.startsWith('Set<');
  bool get isNotSet => !isSet;
  bool get isMap => type == 'Map' || type.startsWith('Map<');
  bool get isNotMap => !isMap;
  bool get isObjectId => type == 'ObjectId';
  bool get isNotObjectId => !isObjectId;
  bool get isPrimitive => isBool || isInt || isNum || isDouble || isString;
  bool get isNotPrimitive => !isPrimitive;
  bool get isBool => type == 'bool';
  bool get isNotBool => !isBool;
  bool get isInt => type == 'int';
  bool get isNotInt => !isInt;
  bool get isNum => type == 'num';
  bool get isNotNum => !isNum;
  bool get isDouble => type == 'double';
  bool get isNotDouble => !isDouble;
  bool get isImported => parent.schema.imports.any((m) => m.names.any((n) => n == type));
  bool get isNotImported => !isImported;
  bool get isInPart => isNotDateTime && isNotCollection && isNotPrimitive && parent.schema.isInPart(type);
  bool get isNotInPart => !isInPart;
  bool get isEnum => isNotDateTime && isNotCollection && isNotPrimitive && parent.schema.isEnum(type);
  bool get isNotEnum => !isEnum;

  bool get isInitialized => initializer != null;
  bool get isNotInitialized => !isInitialized;

  bool get isParameterized => type.contains('<');
  bool get isNotParameterized => !isParameterized;
  String get parameter => type.substring(type.indexOf('<') + 1, type.length-1);

  bool get isRequired => isNotNullable && isNotInitialized;
  bool get isNotRequired => !isRequired;

  String? meta(String key) => metadata.firstWhereOrNull((m) => m.startsWith('$key:'))?.substring('$key:'.length).trim();

  String? get jsonDecode => meta('jsonDecode');
  String? get jsonEncode => meta('jsonEncode');

  String get rawType => type.split('<')[0];
  late String? typeArgument = (){
    final index = type.indexOf('<');
    return (index != -1) ? type.substring(index + 1, type.length - 1) : null;
  }();
  String? get rawTypeArgument {
    final index = type.indexOf('<');
    return (index != -1) ? type.substring(index + 1, type.length - 1).split('<')[0] : null;
  }

  String decoderValue([String? v]) {
    v ??= "d['$name']";
    return parent.schema.imports.firstWhereOrNull((i) => i.decodes(type))?.decoder?.replaceAll('\$', v)
        ?? (isDateTime ? "DateTime.fromMillisecondsSinceEpoch($v)" : null)
        ?? (jsonDecode?.replaceAll('\$', v))
        ?? (isCollection ? decoderCollection(v) : null)
        ?? (isEnum ? '$type.values.byName($v)' : null)
        ?? (isInPart ? '$type.fromJson($v)' : null)
        ?? v;
  }

  String decoderCollection(String value) {
    final parameter = this.parameter;
    if(parameter == null) {
      if(isIterable) return 'List.from($value ?? [])';
      if(isList) return 'List.from($value ?? [])';
      if(isSet) return 'Set.from($value ?? {})';
      if(isMap) return 'Map.from($value ?? {})';
      throw '$this is not a collection';
    }
    final field = ClassField.paramOf(this);
    if(field.isPrimitive) {
      if(isIterable) return 'List<${parameter}>.from($value ?? [])';
      if(isList) return 'List<${parameter}>.from($value ?? [])';
      if(isSet) return 'Set<${parameter}>.from($value ?? {})';
      if(isMap) return 'Map<${parameter}>.from($value ?? {})';
      throw '$this is not a collection';
    }
    final decoder = field.decoderValue('v');
    if(isIterable) return '($value ?? []).map((v) => $decoder).cast<$parameter>()';
    if(isList) return '($value ?? []).map((v) => $decoder).toList().cast<$parameter>()';
    if(isSet) return '($value ?? {}).map((v) => $decoder).toSet().cast<$parameter>()';
    if(isMap) return '($value ?? {}).map((k,v) => MapEntry(\'\$k\', $decoder)).cast<String, $parameter>()';
    throw '$this is not a collection';
  }

  String? get encoderValue {
    final v = "s['$name']";
    String encoder =
      parent.schema.imports.firstWhereOrNull((i) => i.encodes(type))?.encoder?.replaceAll('\$', v)
      ?? jsonEncode?.replaceAll('\$', v)
      ?? v;
    return '($v is $type) ? $encoder : throw InvalidTypeException.value($v, type: $type)';
  }

  String fromString(String v) {
    if(isString) return v;
    if(isInt) return 'int.parse($v)';
    return v;
  }

  @override
  String toString() => '$name $type$nullable';
}
