import 'package:collection/src/iterable_extensions.dart';
import 'package:oobium_datastore_gen/src/class.dart';
import 'package:oobium_datastore_gen/src/class_field.dart';
import 'package:oobium_datastore_gen/src/schema_parser.dart';

class ClassGenerator {

  final Class c;
  ClassGenerator(this.c);
  Schema get schmea => c.schema;
  String get name => c.name;

  String compile() =>
'''
class $name implements Comparable<$name> {
  
  $name($modelParams) : this._($modelConstructor);
  $name.fromJson($fromJsonParams) : this._($fromJsonConstructor);
  static List<$name> fromJsonList($fromJsonListParams) => (list as Iterable).map((data) => $name.fromJson($fromJsonListConstructor)).toList();
  $name._($model_Constructor);
  
  ${c.fields.map((f) => f.fieldDeclaration).join()}
  
  @override
  int compareTo($name other) => $compareToTests;

  $name copyWith($copyWithParams) => $name._($copyWithConstructor);
  
  @override
  bool operator ==(Object? other) => identical(this, other)
    || (other is $name && ($equalsTests));
  
  @override
  int get hashCode => $objectHash;
  
  $toJson
}
''';

  String get model_Constructor => c.fields.map((f) => 'this.${f.name}').join(',');

  String get modelParams => '{'
    '${c.fields.map((f) => '${f.isRequired ? 'required ${f.type}' : '${f.type}${f.nullable}'} ${f.name}${f.isInitialized ? '=${f.initializer}' : ''}').join(',')}'
  '}';

  String get modelConstructor => c.fields.map((f) => f.name).join(',');

  String get compareToTests => [
    ...c.fields.map((f) => 'compare(${f.name}, other.${f.name})'),
    '0',
  ].join('??');

  String get copyWithParams => '{${c.fields.map((f) => [
    '${f.type}? ${f.name}',
    if(f.isNullable) 'bool ${f.name}AsNull=false'
  ]).expand((e) => e).join(',')}}';

  String get copyWithConstructor => c.fields.map((f) => f.isNullable
    ? '${f.name}AsNull ? null : (${f.name} ?? this.${f.name})'
    : '${f.name} ?? this.${f.name}'
  ).join(',');

  String get fromJsonParams => 'Map<String, dynamic> data, {${c.fields.map((f) => [
    '${f.type}? ${f.name}',
    if(f.isNullable) 'bool ${f.name}AsNull=false'
  ]).expand((e) => e).join(',')}}';

  String get fromJsonConstructor => c.fields.map((f) => (f.isNullable || f.isInitialized)
    ? "${f.name} ?? ((data['${f.name}'] != null) ? ${f.jsonDecoder} : ${f.initializer})"
    : "${f.name} ?? ${f.jsonDecoder}"
  ).join(',');

  String get fromJsonListParams => 'list, {${c.fields.map((f) => [
    '${f.type}? ${f.name}',
    if(f.isNullable) 'bool ${f.name}AsNull=false'
  ]).expand((e) => e).join(',')}}';

  String get fromJsonListConstructor => 'data, ${c.fields.map((f) => [
    '${f.name}: ${f.name}',
    if(f.isNullable) '${f.name}AsNull: ${f.name}AsNull'
  ]).expand((e) => e).join(',\n')}';

  String get equalsTests => c.fields.map((f) => '${f.name} == other.${f.name}').join('||');

  String get objectHash => c.fields.length < 20
      ? 'Object.hash($hashValues)' : 'Object.hashAll([$hashValues])';

  String get hashValues => [
    c.name,
    ...c.fields.map((f) => f.name)
  ].join(',');

  String get toJson => 'Map<String, dynamic> toJson() => {${c.fields.map((f) => f.isNullable
    ? "if(${f.name} != null) '${f.name}': ${f.jsonEncoder}"
    : "'${f.name}': ${f.jsonEncoder}").join(',')}'
  '};';
}

extension ClassFieldX on ClassField {

  String get fieldDeclaration => 'final $type$nullable $name;';

  String get jsonDecoder => _jsonDecoder();
  String _jsonDecoder([String? v]) {
    v ??= "data['${name}']";
    return parent.schema.imports.firstWhereOrNull((i) => i.jsonDecodes(type))?.jsonDecoder?.replaceAll('\$', v)
        ?? parent.schema.imports.firstWhereOrNull((i) => i.decodes(type))?.decoder?.replaceAll('\$', v)
        ?? (isDateTime ? 'DateTime.parse($v)' : null)
        ?? (jsonDecode?.replaceAll('\$', v))
        ?? (isCollection ? fromJsonCollection(v) : null)
        ?? (isEnum ? '$type.values.byName($v)' : null)
        ?? (isInPart ? '$type.fromJson($v)' : null)
        ?? v;
  }

  String get jsonEncoder {
    final v = name;
    final e = parent.schema.imports.firstWhereOrNull((i) => i.jsonDecodes(type))?.jsonEncoder?.replaceAll('\$', v)
        ?? parent.schema.imports.firstWhereOrNull((i) => i.decodes(type))?.encoder?.replaceAll('\$', v)
        ?? (jsonEncode?.replaceAll('\$', v))
        ?? (isObjectId ? '$v$nullable.hexString' : null)
        ?? (isDateTime ? '$v$nullable.millisecondsSinceEpoch' : null)
        ?? (isPrimitive ? v : null)
        ?? (isInPart ? '$v$nullable.toJson()' : null)
        ?? (isEnum ? '$v$nullable.name' : null)
        ?? 'jsonValue($v)';
    return isNullable && isInitialized ? '$e ?? $initializer' : e;
  }

  String fromJsonCollection(String value) {
    if(isNotParameterized) {
      if(isIterable) return 'List.from($value ?? [])';
      if(isList) return 'List.from($value ?? [])';
      if(isSet) return 'Set.from($value ?? {})';
      if(isMap) return 'Map.from($value ?? {})';
      throw '$this is not a collection';
    }
    final field = ClassField.paramOf(this);
    if(field.isPrimitive) {
      if(isIterable) return 'List<${parameter}>.from($value ?? [])';
      if(isList) return 'List<${parameter}>.from($value ?? [])';
      if(isSet) return 'Set<${parameter}>.from($value ?? {})';
      if(isMap) return 'Map<${parameter}>.from($value ?? {})';
      throw '$this is not a collection';
    }
    final fromJson = field._jsonDecoder('v');
    if(isIterable) return '($value ?? []).map((v) => $fromJson).cast<$parameter>()';
    if(isList) return '($value ?? []).map((v) => $fromJson).toList().cast<$parameter>()';
    if(isSet) return '($value ?? {}).map((v) => $fromJson).toSet().cast<$parameter>()';
    if(isMap) return '($value ?? {}).map((k,v) => MapEntry(\'\$k\', $fromJson)).cast<String, $parameter>()';
    throw '$this is not a collection';
  }
}