import 'package:oobium_datastore_gen/src/feature.dart';
import 'package:oobium_datastore_gen/src/model.dart';
import 'package:oobium_datastore_gen/src/schema_parser.dart';
import 'package:xstring/xstring.dart';

String genConsole(Schema schema) {
  final dsName = schema.dsName;
  final models = schema.models;
  final fetchable = schema.has<RestFeature>()
    ? models.where((m) => m.canFetch).toList()
    : [];

  return '''
    import 'dart:io';

    import 'package:args/command_runner.dart';
    import '${schema.name}.schema.g.dart';

    mixin _DataMixin {
      Future<$dsName> Function()? _ds;
      Future<$dsName> get ${dsName.varName} => _ds?.call() ?? ((this as Command).parent as _DataMixin).${dsName.varName};
    }

    class ${dsName}Command extends Command with _DataMixin {
      @override String get name => '${dsName.underscored}';
      @override String get description => '${dsName.separated(' ')} commands';
      ${dsName}Command([Future<${dsName}> Function()? ds]) {
        _ds = ds;
        ${fetchable.isNotEmpty ? 'addSubcommand(${dsName}FetchCommand());' : ''}
        addSubcommand(${dsName}GetCommand());
      }
    }
    ${fetchable.isNotEmpty ? '''
    class ${dsName}FetchCommand extends Command with _DataMixin {
      @override String get name => 'fetch';
      @override String get description => '${dsName.separated(' ')} fetch commands';
      ${dsName}FetchCommand() {${models.where((m) => m.canFetch).map((m) =>[
        if(m.isNotSingleton) 'addSubcommand(${dsName}Fetch${m.namePlural}Command());',
        'addSubcommand(${dsName}Fetch${m.name}Command());'].join()).join()}
      }
    }
    ''' : ''}
    class ${dsName}GetCommand extends Command with _DataMixin {
      @override String get name => 'get';
      @override String get description => '${dsName.separated(' ')} get commands';
      ${dsName}GetCommand() {${models.map((m) =>
        m.isNotSingleton ? 'addSubcommand(${dsName}Get${m.namePlural}Command());' : ''
        'addSubcommand(${dsName}Get${m.name}Command());').join()}
      }
    }
    ${fetchable.map((m) => fetch(dsName, m)).join()}
    ${models.map((m) => get(dsName, m)).join()}
  ''';
}

String fetch(String dsName, Model m) => m.isSingleton
  ? singletonFetch(dsName, m)
  : '${pluralFetch(dsName, m)}'
    '${singularFetch(dsName, m)}';

String pluralFetch(String dsName, Model m) => '''
class ${dsName}Fetch${m.namePlural}Command extends Command with _DataMixin {
  @override String get name => '${m.namePlural.underscored}';
  @override String get description => 'fetch ${m.namePlural}';
  @override
  Future<void> run() async {
    await (await ${dsName.varName}).${m.varNamePlural}.fetch();
    stdout.writeln('done');
  }
}
''';

String singularFetch(String dsName, Model m) => '''
class ${dsName}Fetch${m.name}Command extends Command with _DataMixin {
  @override String get name => '${m.name.underscored}';
  @override String get description => 'fetch ${m.name}';
  @override
  Future<void> run() async {
    stdout.write('id: ');
    final s = stdin.readLineSync()?.trim();
    if(s != null) {
      final id = ${m.keyName}.maybe(s);
      if(id != null) {
        await (await ${dsName.varName}).${m.varNamePlural}.fetch(id: id);
        stdout.writeln('done');
      } else {
        stdout.writeln('could not convert \$s to ${m.idType}');
      }
    } else {
      stdout.writeln('canceled');
    }
  }
}
''';

String singletonFetch(String dsName, Model m) => '''
class ${dsName}Fetch${m.name}Command extends Command with _DataMixin {
  @override String get name => '${m.name.underscored}';
  @override String get description => 'fetch ${m.name}';
  @override
  Future<void> run() async {
    await (await ${dsName.varName}).${m.varName}.fetch();
    stdout.writeln('done');
  }
}
''';

String get(String dsName, Model m) => m.isSingleton
  ? singletonGet(dsName, m)
  : '${pluralGet(dsName, m)}'
    '${singularGet(dsName, m)}';

String pluralGet(String dsName, Model m) => '''
class ${dsName}Get${m.namePlural}Command extends Command with _DataMixin {
  @override String get name => '${m.namePlural.underscored}';
  @override String get description => 'get ${m.namePlural}';
  @override
  Future<void> run() async {
    final ${m.varNamePlural} = (await ${dsName.varName}).${m.varNamePlural};
    if(${m.varNamePlural}.isEmpty) {
      stdout.writeln('no results');
    } else {
      for(final ${m.varName} in ${m.varNamePlural}) {
        stdout.writeln(${m.varName}.toModel().toJson());
      }
    }
  }
}
''';

String singularGet(String dsName, Model m) => '''
class ${dsName}Get${m.name}Command extends Command with _DataMixin {
  @override String get name => '${m.name.underscored}';
  @override String get description => 'get ${m.name} by id';
  @override
  Future<void> run() async {
    stdout.write('id: ');
    final s = stdin.readLineSync()?.trim();
    if(s != null) {
      final id = ${m.keyName}.maybe(s);
      if(id != null) {
        final ${m.varName} = (await ${dsName.varName}).${m.varNamePlural}.getOrNull(id);
        if(${m.varName} == null) {
          stdout.writeln('not found');
        } else {
          stdout.writeln(${m.varName}.toModel().toJson());
        }
      } else {
        stdout.writeln('could not convert \$s to ${m.idType}');
      }
    } else {
      stdout.writeln('canceled');
    }
  }
}
''';

String singletonGet(String dsName, Model m) => '''
class ${dsName}Get${m.name}Command extends Command with _DataMixin {
  @override String get name => '${m.name.underscored}';
  @override String get description => 'get ${m.name} by id';
  @override
  Future<void> run() async {
    final ${m.varName} = (await ${dsName.varName}).${m.varName};
    stdout.writeln(${m.varName}.toModel().toJson());
  }
}
''';