import 'package:oobium_datastore_gen/src/model.dart';
import 'package:oobium_datastore_gen/src/model_field.dart';
import 'package:oobium_datastore_gen/src/schema_parser.dart';

class ModelFeatureGenerator {

  final Schema schema;
  ModelFeatureGenerator(this.schema);

  String get dsName => schema.dsName;
  List<Model> get models => schema.models;
  List<Model> get objects => schema.objects;
  List<Model> get types => schema.types;

  String compile() =>
'''
class _ModelFeature extends WatchableStateFeature {
  ${objects.map((m) => m.hasInitialState
    ? 'late final ${m.varName} = (){'
    '  if(!contains(const ${m.keyName}())) {'
    '    put(${m.initialState});'
    '  }'
    '  return ${m.refName}._(this);'
    '}();'
    : 'late final ${m.varName} = ${m.refName}._(this);').join()}
  ${types.map((m) => 'late final ${m.varNamePlural} = ${m.namePlural}(this);').join()}
  void load(data) {${models.map((m) => '''
    if(data is ${m.name}) {
      put(data);
    }''').join('else')}
    else if(data is Iterable) {
      data.forEach((e) => load(e));
    }
    else if(data is Map<String, dynamic>) {
      for(final key in data.keys) {${models.map((m) => '''
        if(key == '${m.varName}' && data[key] is Map<String, dynamic>) {
          load(${m.name}.fromJson(data['${m.varName}']));
        }''').join('else')}${types.map((m) => '''
        else if(key == '${m.varNamePlural}' && data[key] is Iterable) {
          load(${m.name}.fromJsonList(data['${m.varNamePlural}']));
        }''').join()}
        else {
          throw Exception('invalid map entry during load: (\$key, \${data[key]})');
        }
      }    
    }
    else {
      throw Exception('attempted to load unknown type: \$data');
    }
  }
  void unload(data) {${models.map((m) => '''
    if(data is ${m.keyName}) {
      remove(data);
    }''').join('else')}
    else if(data is Iterable) {
      data.forEach((e) => unload(e));
    }
    else if(data is Map<String, dynamic>) {
      for(final key in data.keys) {${models.map((m) => '''
        if(key == '${m.varName}' && data[key] is ${m.idTypeJson}) {
          unload(${m.keyConstructorUnload});
        }''').join('else')}${types.map((m) => '''
        else if(key == '${m.varNamePlural}' && data[key] is Iterable<${m.idTypeJson}>) {
          (data[key] as Iterable).forEach((e) => unload(${m.keyConstructorUnload}));
        }''').join()}
        else {
          throw Exception('invalid map entry during unload: (\$key, \${data[key]})');
        }
      }
    }
    else {
      throw Exception('attempted to unload unknown type: \$data');
    }
  }
  
$onEvents
}
''';

  String get onEvents => models.any((m) => m.hasDependencies) ?
'''
  @override
  void onEvents(StateEvents events) {
    for(final v in events.values) {${models.where((m) => m.hasDependencies).map((m) => '''
      if(v is ${m.name}) {${m.dependenciesFields.map((f) => (f.isNullable)
      ? 'if(v.${f.name} != null) {'
        '  _onRipple(${rippleKey('v.${f.name}!', f)});'
        '}'
      : '_onRipple(${rippleKey('v.${f.name}', f)});').join()}
      }''').join(' else ')}
    }
    super.onEvents(events);
  }
  
  void _onRipple(StateKey k) {
    if(addRipple(k)) {${models.where((m) => m.hasDependencies).map((m) => '''
      if(k is ${m.keyName}) {
        final ${m.varName} = getOrNull(k);${m.dependenciesFields.map((f) => '''
        final ${m.varName}_${f.name} = ${m.varName}?.${f.name};
        if(${m.varName}_${f.name} != null) {
          _onRipple(${rippleKey('${m.varName}_${f.name}', f)});
        }''').join()}
      }''').join(' else ')}
    }
  }
''' : '';

  String rippleKey(String varName, ModelField f) => f.linkedModel!.isSingleton
      ? '${f.linkedModel?.keyName}()'
      : '${f.linkedModel?.keyName}($varName)';
}

extension _ModelGeneratorX on Model {
  String get idTypeJson => isInternal ? 'String' : idType;
  String get keyConstructorUnload => isSingleton ? keyName : isInternal ? '$keyName.parse(data[key])' : '$keyName(data[key] as $idType)';
}