import 'package:collection/collection.dart';
import 'package:oobium_datastore_gen/src/model_field.dart';
import 'package:oobium_datastore_gen/src/schema_parser.dart';
import 'package:xstring/xstring.dart';

class ExternalModel extends Model {

  ExternalModel(Schema schema, String name, String namePlural,
    Iterable<ModelArg>? args,
    Iterable<ModelField>? fields
  ) : super(schema, name, namePlural, args, fields);
  @override bool get isExternal => true;
  @override bool get isInternal => false;
  @override bool get isSingleton => false;
}

class InternalModel extends Model {

  InternalModel(Schema schema, String name, String namePlural,
    Iterable<ModelArg>? args,
    Iterable<ModelField>? fields
  ) : super(schema, name, namePlural, args, [
    ModelField(local: 'id', type: 'ObjectId'),
    ...?fields
  ]);
  @override bool get isExternal => false;
  @override bool get isInternal => true;
  @override bool get isSingleton => false;
}

class Singleton extends Model {

  Singleton(Schema schema, String name,
    Iterable<ModelArg>? args,
    Iterable<ModelField>? fields
  ) : super(schema, name, name, args, fields);
  @override bool get isExternal => false;
  @override bool get isInternal => false;
  @override bool get isSingleton => true;
}

abstract class Model {

  final Schema schema;
  final String name;
  final String namePlural;
  final args = <ModelArg>[];
  final fields = <ModelField>[];

  Model(this.schema, this.name, this.namePlural,
    Iterable<ModelArg>? args,
    Iterable<ModelField>? fields
  ) {
    args?.forEach((a) => this.args.add(a));
    fields?.forEach((f) => this.fields.add(f..model = this));
  }

  late final String keyName = '${name}Key';
  late final String queryName = '${name}Query';
  late final String refName = '${name}Ref';
  late final String varName = name.varName;
  late final String varNamePlural = namePlural.varName;

  bool get isExternal;
  bool get isNotExternal => !isExternal;
  bool get isInternal;
  bool get isNotInternal => !isInternal;
  bool get isSingleton;
  bool get isNotSingleton => !isSingleton;

  bool get canFetch => isSingleton || isExternal;
  bool get cannotFetch => !canFetch;

  bool get canCreate => (isSingleton || isInternal) && dataFields.every((f) => f.isNullable || f.isInitialized);
  bool get cannotBuild => !canCreate;

  Iterable<ModelField> get realFields => fields.where((f) => f.isNotHasMany);
  Iterable<ModelField> get dataFields => fields.where((f) => f.isNotHasMany && f.isNotId);

  ModelField? get idField => fields.firstWhereOrNull((f) => f.name == 'id');
  String get idType => isSingleton ? 'String' : (idField?.type ?? 'ObjectId');
  String get idArg => isSingleton ? '\'$name\'' : 'id';
  String get keyArg => '$keyArgType($idArg)';
  String keyArgMaybe(ModelField f, {bool? nullable}) =>
    isSingleton ? 'const StringKey<$name>(\'$name\')' :
    isInternal ? (nullable ?? f.isNullable) ? 'InternalKey.maybe<$name>(${f.name})' : 'InternalKey<$name>(${f.name})' :
    (idType == 'int') ? (nullable ?? f.isNullable) ? 'IntKey.maybe<$name>(${f.name})' : 'IntKey<$name>(${f.name})' :
    (idType == 'String') ? (nullable ?? f.isNullable) ? 'StringKey.maybe<$name>(${f.name})' : 'StringKey<$name>(${f.name})' :
    throw Exception('invalid type for model id: $idType');
  String get keyArgType =>
    isSingleton ? 'const StringKey<$name>' :
    isInternal ? 'InternalKey<$name>' :
    (idType == 'int') ? 'IntKey<$name>' :
    (idType == 'String') ? 'StringKey<$name>' :
    throw Exception('invalid type for model id: $idType');
  String get keyType => keyName;
    // isSingleton ? 'DataKey<$name, String>' :
    // isInternal ? 'DataKey<$name, ObjectId>' :
    // (idType == 'int') ? 'DataKey<$name, int>' :
    // (idType == 'String') ? 'DataKey<$name, String>' :
    // throw Exception('invalid type for model id: $idType');

  String get newKey => '$keyArgType(${isSingleton ? "'$name'" : ''})';

  String get initialState => '$name._(${[
    if(isExternal) 'id'
    else if(isNotSingleton) 'id ?? ObjectId()',
    ...dataFields.map((f) => f.initializer ?? 'null')
  ].join(',')})';
  bool get hasInitialState => realFields.every((f) => f.isNullable || f.isInitialized);

  /// models that this model depends on
  ///   this model depends on:
  ///     - dependency1
  ///     - dependency2
  ///   defined as any field that isModel
  ///     this model depends on the models linked to in those fields
  ///     (does not need to be a bi-directional link)
  late final dependencies = dependenciesFields.map((f) => f.typeModel!).toSet().toList();
  late final dependenciesFields = fields.where((f) => f.isModel).toList();
  bool get hasDependencies => isDependent;
  bool get isDependent => dependenciesFields.isNotEmpty;
  bool get isNotDependent => !isDependent;

  /// models that depend on this model
  ///   dependency1 depends on:
  ///     - this
  ///   dependency2 depends on:
  ///     - this
  ///   defined as any field that is pointed to by another field that isModel
  ///     this model is a model that another model depends on, through a link in its fields
  ///     (does not need to be a bi-directional link)
  late final dependentsFields = schema.models.map((m) => m.fields.where((f) => f.typeModel == this)).expand((e) => e).toList();
  late final dependents = dependentsFields.map((f) => f.model).toSet().toList();
  bool get hasDependents => dependents.isNotEmpty;
  bool get isDependency => hasDependents;
  bool get isNotDependency => !isDependency;

  bool get isGeneric => name.contains('<');
  bool get isNotGeneric => !isGeneric;

  String? get typeParameter {
    final index = name.indexOf('<');
    return (index != -1) ? name.substring(index + 1, name.length - 1) : null;
  }
  String? get typeArgument {
    if(isGeneric) {
      return null;
    }
    final index = name.indexOf('<');
    return (index != -1) ? name.substring(index + 1, name.length - 1) : null;
  }

  @override
  String toString() => name;
}

class ModelArg {
  final String name;
  final String value;
  ModelArg({required this.name, required this.value});
}
