import 'package:collection/collection.dart';
import 'package:oobium_datastore_gen/src/class.dart';
import 'package:oobium_datastore_gen/src/model.dart';

class ModelField {

  final String local;
  final String remote;
  final String type;
  final String? _hasManyField;
  final bool _nullable;
  final String? initializer;
  final metadata = <String>[];

  ModelField({
    required this.local,
    String? remote,
    required this.type,
    String? hasManyField,
    bool nullable = false,
    String? initializer,
    Iterable<String>? metadata
  }) :
    remote = remote ?? local,
    _hasManyField = hasManyField,
    _nullable = nullable,
    initializer = initializer {
    metadata?.forEach((m) => this.metadata.add(m));
  }
  ModelField.paramOf(ModelField f) :
        model = f.model, type = f.parameter, _hasManyField = f._hasManyField, local ='', remote = '', _nullable = false, initializer = null;

  late final Model model;
  String get name => local;
  String get jsonField => remote;
  late final Class? typeClass = model.schema.classes.firstWhereOrNull((m) => m.name == rawType);
  late final Model? typeModel = model.schema.models.firstWhereOrNull((m) => m.name == rawType);

  String get keyType => isModel ? '${type}Key' : type;
  String get refType => isModel ? '${type}Ref' : type;
  String get idType => typeModel?.idType ?? (throw Exception('called idType on field that isNotModel'));

  String get nullable => isNullable ? '?' : '';

  bool get isId => name == 'id';
  bool get isNotId => !isId;
  bool get isDateTime => type == 'DateTime';
  bool get isNotDateTime => !isDateTime;
  bool get isNullable => _nullable;
  bool get isNotNullable => !isNullable;
  bool get isString => type == 'String';
  bool get isNotString => !isString;
  bool get isIterable => type == 'Iterable' || type.startsWith('Iterable<');
  bool get isNotIterable => !isIterable;
  bool get isCollection => isList || isIterable || isSet || isMap;
  bool get isNotCollection => !isCollection;
  bool get isList => type == 'List' || type.startsWith('List<');
  bool get isNotList => !isList;
  bool get isSet => type == 'Set' || type.startsWith('Set<');
  bool get isNotSet => !isSet;
  bool get isMap => type == 'Map' || type.startsWith('Map<');
  bool get isNotMap => !isMap;
  bool get isObjectId => type == 'ObjectId';
  bool get isNotObjectId => !isObjectId;
  bool get isPrimitive => isBool || isInt || isNum || isDouble || isString;
  bool get isNotPrimitive => !isPrimitive;
  bool get isBool => type == 'bool';
  bool get isNotBool => !isBool;
  bool get isInt => type == 'int';
  bool get isNotInt => !isInt;
  bool get isNum => type == 'num';
  bool get isNotNum => !isNum;
  bool get isDouble => type == 'double';
  bool get isNotDouble => !isDouble;
  bool get isClass => typeClass != null;
  bool get isNotClass => !isClass;
  bool get isModel => typeModel != null;
  bool get isNotModel => !isModel;
  bool get isImported => model.schema.imports.any((m) => m.names.any((n) => n == type));
  bool get isNotImported => !isImported;
  bool get isInPart => isNotDateTime && isNotModel && isNotCollection && isNotPrimitive && model.schema.isInPart(type);
  bool get isNotInPart => !isInPart;
  bool get isEnum => isNotDateTime && isNotModel && isNotCollection && isNotPrimitive && model.schema.isEnum(type);
  bool get isNotEnum => !isEnum;
  bool get isHasMany => isIterable && _hasManyField != null;
  bool get isNotHasMany => !isHasMany;

  late final Model? linkedModel =
    isModel ? typeModel! :
    isHasMany ? model.schema.models.firstWhereOrNull((m) => m.name == typeArgument) :
    null;
  late final ModelField? linkedField =
    isModel ? linkedModel?.fields.firstWhereOrNull((f) => name == f._hasManyField) :
    isHasMany ? linkedModel?.fields.firstWhereOrNull((f) => f.name == _hasManyField) :
    null;
  bool get isLinked => linkedField != null;
  bool get isNotLinked => !isLinked;

  bool get isOverride => isId;

  bool get isRequired => isNotNullable && isNotInitialized;
  bool get isNotRequired => !isRequired;

  bool get isInitialized => initializer != null;
  bool get isNotInitialized => !isInitialized;

  bool get isGeneric => model.isGeneric && rawTypeArgument == model.typeParameter;
  bool get isNotGeneric => !isGeneric;

  bool get isParameterized => isNotGeneric && type.contains('<');
  bool get isNotParameterized => !isParameterized;
  String get parameter => type.substring(type.indexOf('<') + 1, type.length-1);

  String? meta(String key) => metadata.firstWhereOrNull((m) => m.startsWith('$key:'))?.substring('$key:'.length).trim();

  String? get jsonDecode => meta('jsonDecode');
  String? get jsonEncode => meta('jsonEncode');

  String get rawType => type.split('<')[0];
  String? get rawTypeArgument {
    final index = type.indexOf('<');
    return (index != -1) ? type.substring(index + 1, type.length - 1).split('<')[0] : null;
  }

  late final String? typeArgument = (){
    if(isGeneric) return null;
    final index = type.indexOf('<');
    return (index != -1) ? type.substring(index + 1, type.length - 1) : null;
  }();

  @override
  String toString() => '$local:$remote $type$nullable${(_hasManyField != null) ? ':$_hasManyField' : ''}';
}
