import 'package:collection/collection.dart';
import 'package:oobium_datastore_gen/src/class.dart';
import 'package:oobium_datastore_gen/src/class_field.dart';
import 'package:oobium_datastore_gen/src/feature.dart';
import 'package:oobium_datastore_gen/src/model.dart';
import 'package:oobium_datastore_gen/src/model_field.dart';
import 'package:oobium_datastore_gen/src/util.dart';
import 'package:xstring/xstring.dart';

class SchemaParser {

  Schema parse(String name, String? version, Iterable<String> lines) {
    final elements = SchemaElements.load(lines);
    final schema = Schema(name, version);
    schema.features.addAll(features(elements));
    schema.imports.addAll(elements._imports);
    schema.classes.addAll(classes(schema, elements));
    schema.models.addAll(models(schema, elements));
    if(schema.models.any((m) => m.hasDependents)) {
      schema.features.add(Feature.trigger());
    }
    // TODO unknown elements
    return schema;
  }

  List<Feature> features(SchemaElements elements) {
    return elements._features.map((f) =>
      Feature.create(f.name, f.options)
    ).toList();
  }

  List<Class> classes(Schema schema, SchemaElements elements) {
    return elements._classes.map((c) {
      return Class(
        schema: schema,
        name: c.name,
        singleton: c.singleton,
        fields: c.fields.map((f) => ClassField(
          name: f.local,
          type: f.type,
          nullable: f.nullable,
          initializer: f.initializer,
          metadata: f.options,
        ))
      );
    }).toList();
  }

  List<Model> models(Schema schema, SchemaElements elements) {
    return elements._models.map((m) {
      final args = m.args.map((a) => ModelArg(
        name: a.name,
        value: a.value,
      ));
      final fields = m.fields.map((f) => ModelField(
        local: f.local,
        remote: f.remote,
        type: f.type,
        hasManyField: f.key,
        nullable: f.nullable,
        initializer: f.initializer,
        metadata: f.options,
      ));
      if(m.singleton) {
        return Singleton(schema, m.name, args, fields);
      }
      if(fields.any((f) => f.isId)) {
        return ExternalModel(schema, m.name, m.namePlural, args, fields);
      }
      return InternalModel(schema, m.name, m.namePlural, args, fields);
    }).toList();
  }
}

class Schema {

  final String name;
  final String? version;
  final imports = <SchemaImport>[];
  final classes = <Class>[];
  final models = <Model>[];
  final features = <Feature>[];
  Schema(this.name, this.version);

  bool has<T extends Feature>() => features.any((f) => f is T);

  late bool Function(dynamic type) isEnum;
  late bool Function(dynamic type) isInPart;

  bool get isVersioned => version != null;
  bool get isNotVersioned => !isVersioned;

  String get dsName => '${name.camelCase}Data';
  String get dsPath => name.underscored;

  String get partOfDeclaration => "part of '${[
    name,
    if(isVersioned) version,
    'schema.g.dart'
  ].join('.')}';";

  List<Model> get objects => models.where((m) => m.isSingleton).toList();
  List<Model> get types => models.where((m) => m.isNotSingleton).toList();
}

class SchemaElements {

  final _features = <SchemaFeature>[];
  final _imports = <SchemaImport>[];
  final _models = <SchemaModel>[];
  final _classes = <SchemaClass>[];
  SchemaElements();

  SchemaModel? getModel(String name) => _models.firstWhereOrNull((e) => e.isModel && e.name == name);
  bool hasModel(String name) => _models.any((e) => e.isModel && e.name == name);

  static SchemaElements load(Iterable<String> lines) {
    final library = SchemaElements();

    List<int> indents = [];
    SchemaImport? import;
    SchemaClass? clazz;
    SchemaModel? model;
    List<SchemaField>? fields;
    List<String>? options;
    for(final line in lines.filtered) {
      final start = line.indexOf(RegExp(r'\S'));
      if(start == -1) { // blank lines do nothing
        continue;
      }
      else if(start == 0) { // new element
        indents = [0];
        clazz = null;
        model = null;
        import = null;
        fields = null;
        options = null;
        if(line.startsWith('features ')) {
          final features = line.substring(9)
              .split(RegExp(r'\s*,\s*')).map((e) => e.split(':'));
          library._features.addAll(features.map((parts) =>
              SchemaFeature(parts[0], parts.sublist(1))
          ));
        }
        else if(line.startsWith('import ')) {
          if(line.length > 7) {
            final names = line.substring(7).split(RegExp(r'\s*,\s*'));
            library._imports.add(
              import = SchemaImport(library, names)
            );
          }
        }
        else if(line.startsWith('data ')) {
          final matches = RegExp(r'data\s+(\w+)\s*(\([\w:, ]*\))?').firstMatch(line);
          if(matches == null) {
            throw Exception('could not parse model line: $line');
          } else {
            final name = matches.group(1)!;
            final singleton = matches.group(2) != null;
            library._classes.add(
                clazz = SchemaClass(library, name, singleton)
            );
          }
        }
        else {
          final matches = RegExp(r'\s*(\w+)(/\w+)?\s*(\([\w:, ]*\))?').firstMatch(line);
          if(matches == null) {
            throw Exception('could not parse model line: $line');
          } else {
            final name = matches.group(1)!;
            final plural = matches.group(2)?.substring(1);
            final singleton = matches.group(3) != null;
            final options = matches.group(3)?.slice(1, -1) // TODO obsolete?
                .split(RegExp(r',\s*'))
                .where((s) => s.isNotBlank)
                .map((s) => s.trim().split(RegExp(r'\s*:\s*')))
                .map((sa) => SchemaModelArg(sa[0], sa[1])).toList()
                ?? [];
            if(singleton && plural != null) {
              throw Exception('singletons cannot have plurals ($line)');
            }
            final namePlural = plural ?? pluralize(name);
            library._models.add(
              model = SchemaModel(library, name, namePlural, singleton, options)
            );
          }
        }
      }
      else if(start == indents.last) { // another of whatever the last was
        if(options != null) {
          addOption(line, options);
        }
        else if(fields != null) {
          addField(line, fields);
        }
        else {
          throw '? at line: $line';
        }
      }
      else if(start > indents.last) { // new element, field or option
        indents.add(start);
        if(options != null) {
          throw 'invalid indent, line: $line';
        }
        else if(fields != null) {
          options = addOption(line, fields.last.options);
        }
        else if(clazz != null) {
          fields = addField(line, clazz.fields);
        }
        else if(model != null) {
          fields = addField(line, model._fields);
        }
        else if(import != null) {
          options = addOption(line, import.options);
        }
      }
      else { // start < indent : finish, parse if line is not blank
        while(indents.isNotEmpty && start < indents.last) {
          indents.removeLast();
          if(options != null) {
            options = null;
          }
          else if(fields != null) {
            fields = null;
          }
          if(indents.isNotEmpty && start == indents.last) {
            if(fields != null) {
              addField(line, fields);
            }
          }
        }
      }
    }
    return library;
  }

  static List<SchemaField> addField(String line, List<SchemaField> fields) {
    final matches = RegExp(r"\s+(\w+)(\:\w+)?\s+([<\w, :>]+)(\?)?(=.+)?").firstMatch(line);
    if(matches != null) {
      final local = matches.group(1)!;
      final remote = matches.group(2)?.substring(1) ?? local;
      final g3 = matches.group(3)!;
      final link = (g3.startsWith('Iterable') && g3.contains(':'));
      final key = link ? g3.slice(g3.indexOf(':')+1, -1).trim() : null;
      final type = link ? '${g3.substring(0, g3.indexOf(':'))}>' : g3;
      final nullable = matches.group(4) == '?';
      final initializer = matches.group(5)?.substring(1).trim();
      fields.add(SchemaField(local, remote, type, key, nullable, initializer));
    }
    return fields;
  }

  static List<String> addOption(String line, List<String> options) {
    return options..add(line.trim().replaceFirst(' ', ':'));
  }
}

class SchemaFeature {
  final String name;
  final List<String> options;
  SchemaFeature(this.name, this.options);
}

class SchemaImport {
  final SchemaElements library;
  final List<String> names;
  final options = <String>[];

  SchemaImport(this.library, this.names);

  String get name => names.first;
  String? get from => value('from');
  String? get decoder => value('decode');
  String? get encoder => value('encode');
  String? get jsonDecoder => value('jsonDecode');
  String? get jsonEncoder => value('jsonEncode');

  String get package {
    final from = this.from;
    if(from != null && from.startsWith('package:')) {
      return from.substring(8, from.indexOf('/'));
    }
    return '';
  }
  String get path {
    final from = this.from;
    if(from != null) {
      if(from.startsWith('package:')) {
        return 'lib/${from.substring(from.indexOf('/'))}';
      }
      throw 'non-package paths not yet implemented ($from)';
    }
    return '';
  }

  bool has(String name) => options.any((o) => o.startsWith('$name:'));

  String? value(String name) => options
      .firstWhereOrNull((o) => o.startsWith('$name:'))
      ?.substring('$name:'.length);

  bool decodes(String type) => (name == type) && has('decode');
  bool encodes(String type) => (name == type) && has('encode');

  bool jsonDecodes(String type) => (name == type) && has('jsonDecode');
  bool jsonEncodes(String type) => (name == type) && has('jsonEncode');
}

class SchemaClass {
  final SchemaElements library;
  final String name;
  final bool singleton;
  final fields = <SchemaField>[];

  SchemaClass(this.library, this.name, this.singleton);

  bool get isExtension => RegExp(r'^[a-z].*$').hasMatch(name);
  bool get isNotExtension => !isExtension;

  bool get isModel => RegExp(r'^[A-Z].*$').hasMatch(name);
  bool get isNotModel => !isModel;

  @override
  String toString() => '$name()${fields.isNotEmpty ? '\n  ${fields.join('\n ')}' : ''}';
}

class SchemaModel {

  final SchemaElements library;
  final String name;
  final String namePlural;
  final bool singleton;
  final List<SchemaModelArg> args;
  final _fields = <SchemaField>[];

  SchemaModel(this.library, this.name, this.namePlural, this.singleton, this.args);

  // TODO obsolete?
  List<SchemaField> get fields {
    List<SchemaField> fields = [];
    fields.addAll(_fields);
    for(var option in args) {
      final extension = library._models.firstWhereOrNull((e) => e.name == option);
      if(extension != null) {
        fields.addAll(extension.fields);
      }
    }
    return fields;
  }

  bool get isExtension => RegExp(r'^[a-z].*$').hasMatch(name);
  bool get isNotExtension => !isExtension;

  bool get isModel => RegExp(r'^[A-Z].*$').hasMatch(name);
  bool get isNotModel => !isModel;

  @override
  String toString() => '$name(${args.join(', ')})${fields.isNotEmpty ? '\n  ${fields.join('\n ')}' : ''}';
}

class SchemaModelArg {
  final String name;
  final String value;
  SchemaModelArg(this.name, this.value);
}

class SchemaField {
  final String local;
  final String remote;
  final String type;
  final String? key;
  final bool nullable;
  final String? initializer;
  final options = <String>[];
  SchemaField(this.local, this.remote, this.type, this.key, this.nullable, this.initializer);

  @override
  String toString() => '$local $type(${options.join(', ')})';
}

extension XStringIterable on Iterable<String> {
  Iterable<String> get filtered => this.map((line) => line.split('//')[0].trimRight());
}