import 'package:oobium_datastore_gen/src/feature.dart';
import 'package:oobium_datastore_gen/src/schema_parser.dart';
import 'package:test/test.dart';

void main() {
  test('parse features', () {
    final schema = SchemaParser().parse('test', null, ['features console, persist, rest']);
    expect(schema.name, 'test');
    expect(schema.features.length, 3);
    expect(schema.features[0], isA<ConsoleFeature>());
    expect(schema.features[1], isA<PersistFeature>());
    expect(schema.features[2], isA<RestFeature>());
    expect(schema.models, isEmpty);
  });
  test('parse feature with version', () {
    final schema = SchemaParser().parse('test', null, ['features console, persist:123, rest']);
    expect(schema.name, 'test');
    expect(schema.isNotVersioned, isTrue);
    expect(schema.features.whereType<PersistFeature>().first.version, 123);
    expect(schema.models, isEmpty);
  });
  test('parse feature with invalid version', () {
    final schema = SchemaParser().parse('test', null, ['features console, persist:bob, rest']);
    expect(schema.name, 'test');
    expect(schema.isNotVersioned, isTrue);
    expect(schema.features.whereType<PersistFeature>().first.version, 1);
    expect(schema.models, isEmpty);
  });
  test('parse one empty model as external', () {
    final schema = SchemaParser().parse('test', null, ['Person']);
    expect(schema.name, 'test');
    expect(schema.models.length, 1);
    expect(schema.models.single.isExternal, isFalse);
    expect(schema.models.single.isInternal, isTrue);
    expect(schema.models.single.isSingleton, isFalse);
    expect(schema.models.single.name, 'Person');
    expect(schema.models.single.namePlural, 'Persons');
    expect(schema.models.single.args, isEmpty);
  });
  test('parse model as singleton', () {
    final schema = SchemaParser().parse('test', null, ['Person()']);
    expect(schema.name, 'test');
    expect(schema.models.length, 1);
    expect(schema.models.single.isExternal, isFalse);
    expect(schema.models.single.isInternal, isFalse);
    expect(schema.models.single.isSingleton, isTrue);
    expect(schema.models.single.idType, 'String');
    expect(schema.models.single.name, 'Person');
    expect(schema.models.single.args, isEmpty);
  });
  test('parse model as singleton with options', () {
    final schema = SchemaParser().parse('test', null, ['Person(transient: true)']);
    expect(schema.name, 'test');
    expect(schema.models.length, 1);
    expect(schema.models.single.isExternal, isFalse);
    expect(schema.models.single.isInternal, isFalse);
    expect(schema.models.single.isSingleton, isTrue);
    expect(schema.models.single.name, 'Person');
    expect(schema.models.single.args.single.name, 'transient');
    expect(schema.models.single.args.single.value, 'true');
  });
  test('parse provided plural', () {
    final schema = SchemaParser().parse('test', null, ['Person/People']);
    expect(schema.models.single.name, 'Person');
    expect(schema.models.single.namePlural, 'People');
    expect(schema.models.single.args, isEmpty);
  });
  test('parse provided plural crashes singleton', () {
    expect(() => SchemaParser().parse('test', null, ['Person/People()']), throwsException);
  });
  test('embedded iterable', () {
    final schema = SchemaParser().parse('test', null, [
      'Parent',
      '  children Iterable<Child>',
      'Child',
      '  parent Parent',
    ]);
    expect(schema.models.length, 2);
    final parent = schema.models.firstWhere((m) => m.name == 'Parent');
    final child = schema.models.firstWhere((m) => m.name == 'Child');
    final parentField = parent.fields.where((f) => f.isNotId).single;
    final childField = child.fields.where((f) => f.isNotId).single;
    expect(parentField.name, 'children');
    expect(parentField.type, 'Iterable<Child>');
    expect(parentField.typeModel, isNull);
    expect(parentField.isHasMany, isFalse);
    expect(childField.name, 'parent');
    expect(childField.type, 'Parent');
    expect(childField.typeModel, parent);
    expect(childField.isModel, isTrue);
  });
  test('models with hasMany relationship', () {
    final schema = SchemaParser().parse('test', null, [
      'Multiple',
      '  many Iterable<Single:one>',
      'Single',
      '  one Multiple',
    ]);
    expect(schema.models.length, 2);
    final multi = schema.models.firstWhere((m) => m.name == 'Multiple');
    final single = schema.models.firstWhere((m) => m.name == 'Single');
    final multiField = multi.fields.where((f) => f.isNotId).single;
    final singleField = single.fields.where((f) => f.isNotId).single;
    expect(multiField.name, 'many');
    expect(multiField.type, 'Iterable<Single>');
    expect(multiField.typeModel, isNull);
    expect(multiField.isLinked, isTrue);
    expect(multiField.linkedModel, single);
    expect(multiField.isHasMany, isTrue);
    expect(multi.dependents.length, 1);
    expect(multi.dependents.first, single);
    expect(multi.dependentsFields.length, 1);
    expect(multi.dependentsFields.first.name, 'one');
    expect(multi.dependentsFields.first.model, single);
    expect(multi.dependentsFields.first.linkedModel, multi);
    expect(multi.dependencies, isEmpty);
    expect(singleField.name, 'one');
    expect(singleField.type, 'Multiple');
    expect(singleField.typeModel, multi);
    expect(singleField.isModel, isTrue);
    expect(single.dependents, isEmpty);
    expect(single.dependentsFields, isEmpty);
    expect(single.dependencies.length, 1);
    expect(single.dependencies.first, multi);
    expect(single.dependenciesFields.length, 1);
    expect(single.dependenciesFields.first.name, 'one');
    expect(single.dependenciesFields.first.model, single);
    expect(single.dependenciesFields.first.linkedModel, multi);

    expect(multi.dependentsFields.first, single.dependenciesFields.first);
  });
  test('parse data class', () {
    final schema = SchemaParser().parse('test', null, ['data Address']);
    expect(schema.classes.length, 1);
    expect(schema.classes.single.name, 'Address');
    expect(schema.classes.single.singleton, isFalse);
  });
  test('parse singleton data class', () {
    final schema = SchemaParser().parse('test', null, ['data Address()']);
    expect(schema.classes.length, 1);
    expect(schema.classes.single.name, 'Address');
    expect(schema.classes.single.singleton, isTrue);
  });
  test('parse data class fields', () {
    final schema = SchemaParser().parse('test', null, [
      'data Address',
      '  street1 String',
      '  street2 String?',
    ]);
    expect(schema.classes.length, 1);
    expect(schema.classes.single.name, 'Address');
    expect(schema.classes.single.singleton, isFalse);
    expect(schema.classes.single.fields.length, 2);
    expect(schema.classes.single.fields[0].name, 'street1');
    expect(schema.classes.single.fields[0].type, 'String');
    expect(schema.classes.single.fields[0].isNullable, isFalse);
    expect(schema.classes.single.fields[1].name, 'street2');
    expect(schema.classes.single.fields[1].type, 'String');
    expect(schema.classes.single.fields[1].isNullable, isTrue);
  });
}